/* 
 * File   : ffield.cpp
 * Version: 2008-06-06
 *
 * Copyright (C) 2008 Jochen Peters
 *
 * dev@emirp.net
 * http://www.emirp.net/
 * 
 * This file is part of FissionField and FissionField2x.
 *
 * FissionField(2x) is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * FissionField(2x) is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with FissionField(2x). If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <iostream>
#include "ffield.hpp"

namespace ffield {
	
	/*
	 * Implementation of class Atom
	 */
	
	Atom::Atom(int maxNumParticles) {
	
		this->maxNumParticles = maxNumParticles;
		this->numParticles = 0;
		this->team = NONE;
	
	}
	
	void Atom::reset() {
		numParticles = 0;
		this->team = NONE;
	}

	int Atom::getNumParticles() const {
		return numParticles;
	}
	
	int Atom::getMaxNumParticles() const {
		return maxNumParticles;
	}
	
	enum color Atom::getTeam() const {
		return team;
	}
	
	int Atom::addParticle(enum color team) {
		
		int retval;

		/* Add particle */
		numParticles++;
		this->team = team;
		
		if(numParticles < maxNumParticles) {
			retval = 0;
		}
		else {
			this->reset();
			retval = maxNumParticles;
		}
	
		return retval;
	
	}
	
	
	/*
	 * Implementation of class AtomicField
	 */
	
	AtomicField::AtomicField(int size) {
	
		if(size < 2) {
			exit(1);
		}
		
		this->size = size;
		this->numMoves = 0;
		this->score[RED] = 0;
		this->score[BLUE] = 0;
		
		atoms = new Atom**[size];
	
		for(int i=0; i<size; i++) {
			atoms[i] = new Atom*[size];
		}
	
		for(int i=0; i<size; i++) {
			for(int j=0; j<size; j++) {
				if(i==0 || j==0 || i==(size-1) || j==(size-1)) {
					if(i==0 && j==0 || i==0 && j==(size-1) || i==(size-1) && j==0 || i==(size-1) && j==(size-1)) { // Corners
						atoms[i][j] = new Atom(2);
					}
					else { // Sides
						atoms[i][j] = new Atom(3);
					}
				}
				else { // Inside of the field
					atoms[i][j] = new Atom(4);
				}
			}
		}
	
	}
	
	AtomicField::~AtomicField() {
	
		for(int i=0; i<size; i++) {
			for(int j=0; j<size; j++) {
				delete atoms[i][j];
			}
			delete[] atoms[i];
		}
		delete[] atoms;
	
	}
	
	AtomicField::AtomicField(const AtomicField& field) {

		this->size = field.size;
		this->numMoves = field.numMoves;
		this->score[RED] = field.score[RED];
		this->score[BLUE] = field.score[BLUE];
		
		/* Allocate memory and copy atoms */
		atoms = new Atom**[size];
		for(int i=0; i<size; i++) {
			atoms[i] = new Atom*[size];
			for(int j=0; j<size; j++) {
				atoms[i][j] = new Atom(field.atoms[i][j]->getMaxNumParticles());
				*atoms[i][j] = *field.atoms[i][j]; // copy
			}
		}

	}

	Atom* AtomicField::getAtom(int row, int col) const {
		return atoms[row][col];
	}
	
	int AtomicField::getSize() const {
		return size;
	}

	int AtomicField::getNumMoves() const {
		return numMoves;
	}
	
	bool AtomicField::validCoords(int row, int col) const {
	
		if(row < 0 || row >= size || col < 0 || col >= size) {
			return false;
		}
		else {
			return true;
		}
	
	}
	
	bool AtomicField::okayToAdd(int row, int col, enum color team) const {
		
		if(!validCoords(row, col) || team == NONE) {
			return false;
		}
	
		enum color teamOfAtom = atoms[row][col]->getTeam();
	
		if(teamOfAtom == NONE || teamOfAtom == team) {
			return true;
		}
		else {
			return false;
		}
	
	}
	
	int AtomicField::addParticle(int row, int col, enum color team) {
	
		if(!validCoords(row, col)) {
			return -1;
		}
	
		/* If the moving team has a score of 1, the added particle may result in a temporary score of 0
		   and lead to an ionvalid victory of the other team. So we save the score of the moving team now
		   and check it later. */
		int preScore = score[team];
	
		/* Add particle */
		int explode = atoms[row][col]->addParticle(team);
		addUpdate(row, col); // visual update
	
		/* Check if there is a winner */
		if(checkForWinner() != NONE && preScore > 1) {
			return -1;
		}
	
		if(explode > 0) {
			
			/* Spread particles (recursive function calls) */
			explodeUpdate(row, col, team, 4); // visualize the explosion
			addParticle(row-1, col, team);
			explodeUpdate(row, col, team, 3); // visualize the explosion
			addParticle(row, col+1, team);
			explodeUpdate(row, col, team, 2); // visualize the explosion
			addParticle(row+1, col, team);
			explodeUpdate(row, col, team, 1); // visualize the explosion
			addParticle(row, col-1, team);
			explodeUpdate(row, col, team, 0); // visualize the explosion
	
		}
	
		return explode;
		
	}
	
	bool AtomicField::makeMove(int row, int col, enum color team) {
		
		if(!okayToAdd(row, col, team)) {
			return false;
		}
	
		addParticle(row, col, team);
		numMoves++;
	
		return true;
	
	}
	
	int AtomicField::getScore(enum color team) const {
		return score[team];
	}

	int AtomicField::getNumParticles(enum color team) const {
	
		int numParticles = 0;
	
		for(int i=0; i<size; i++) {
			for(int j=0; j<size; j++) {
				enum color actualTeam = atoms[i][j]->getTeam();
				if(actualTeam == team) {
					numParticles += atoms[i][j]->getNumParticles();
				}
			}
		}

		return numParticles;
	}
	
	enum color AtomicField::checkForWinner() {
	
		updateScore();
	
		if(numMoves <= 1) {
			return NONE;
		}
	
		if(score[RED] > 0 && score[BLUE] == 0) {
			return RED;
		}
		else if(score[RED] == 0 && score[BLUE] > 0) {
			return BLUE;
		}
		else {
			return NONE;
		}
		
	}
	
	void AtomicField::updateScore() {
	
		score[RED] = 0;
		score[BLUE] = 0;
	
		for(int i=0; i<size; i++) {
			for(int j=0; j<size; j++) {
				enum color team = atoms[i][j]->getTeam();
				if(team != NONE) {
					score[team]++;
				}
			}
		}
		
	}
	
	void AtomicField::printField() const {
	
		std::cout << "\n";
		
		/* Print column numbers */
		std::cout << "  ";
		for(int i=0; i<size; i++) {
			std::cout << "  " << i+1 << "  ";
		}
		std::cout << "\n";
	
		/* Print line */
		std::cout << "  ";
		for(int i=0; i<size; i++) {
			std::cout << "-----";
		}
		std::cout << "-\n";
	
		for(int i=0; i<size; i++) {
	
			/* Print row numbers */
			std::cout << i+1 << " |";
	
			/* Print field */
			for(int j=0; j<size; j++) {
				enum color team = atoms[i][j]->getTeam();
				if(team == RED) {
					std::cout << " R";
				}
				else if(team == BLUE) {
					std::cout << " B";
				}
				else {
					std::cout << "  ";
				}
	
				int num = atoms[i][j]->getNumParticles();
	
				if(num > 0) {
					std::cout << num << " |";
				}
				else {
					std::cout << "  |";
				}
			}
			std::cout << "\n";
	
			/* Print line */
			std::cout << "  ";
			for(int j=0; j<size; j++) {
				std::cout << "-----";
			}
			std::cout << "-\n";
	
		}
	
		/* Print actual score */
		std::cout << "  RED " << score[RED] << " - BLUE " << score[BLUE] << "\n\n";
	
	}
	
	void AtomicField::addUpdate(int row, int col) {
		// do nothing (meant to be overridden)
	}
	
	
	void AtomicField::explodeUpdate(int row, int col, enum color team, int phase) {
		// do nothing (meant to be overridden)
	}
	
}
