/* 
 * File   : ffield_player.cpp
 * Version: 2008-06-06
 *
 * Copyright (C) 2008 Jochen Peters
 *
 * dev@emirp.net
 * http://www.emirp.net/
 * 
 * This file is part of FissionField and FissionField2x.
 *
 * FissionField(2x) is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * FissionField(2x) is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with FissionField(2x). If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <iostream>
#include <time.h>
#include "ffield_player.hpp"

/* Floating point stuff */
#define EPSILON 0.01
#define FABS(x) ((x) >= 0.0 ? (x) : -(x))
#define FISEQUAL(a,b) (FABS((a) - (b)) <= EPSILON)
#define FUNEQUAL(a,b) (FABS((a) - (b)) > EPSILON)

namespace ffield {
	
	/*
	 * Implementation of class AtomvsPlayer
	 */

	AtomvsPlayer::AtomvsPlayer(enum color team) {
		this->team = team;
	}

	AtomvsPlayer::~AtomvsPlayer() { }

	int AtomvsPlayer::getNumValidMoves(const AtomicField* actualField) {

		/* Count valid moves */
		int numValidMoves = 0;
		for(int i=0; i<actualField->getSize(); i++) {
			for(int j=0; j<actualField->getSize(); j++) {
				if(actualField->okayToAdd(i, j, team)) {
					numValidMoves++;
				}
			}
		}

		return numValidMoves;
	}

	void AtomvsPlayer::getValidMoves(const AtomicField* actualField, int* validMoves_row, int* validMoves_col) {

		int moveIndex = 0;
		for(int i=0; i<actualField->getSize(); i++) {
			for(int j=0; j<actualField->getSize(); j++) {
				if(actualField->okayToAdd(i, j, team)) {
					validMoves_row[moveIndex] = i;
					validMoves_col[moveIndex] = j;
					moveIndex++;
				}
			}
		}
	}


	/*
	 * Implementation of class EvalPlayer
	 */

	EvalPlayer::EvalPlayer(enum color team) : AtomvsPlayer(team) {
		srand((unsigned) time(NULL)); // initialize random number generator
	}

	void EvalPlayer::makeMove(const AtomicField* actualField, int* moveToRow, int* moveToCol) {

		/* Get valid moves */
		int numValidMoves = getNumValidMoves(actualField);
		int* validMoves_row = new int[numValidMoves];
		int* validMoves_col = new int[numValidMoves];
		getValidMoves(actualField, validMoves_row, validMoves_col);

		/* Determine best moves */
		float bestScore = -1;
		int numBestMoves = 0;
		int bestMovesIndex = 0;
		int* bestMoves_row = new int[numValidMoves];
		int* bestMoves_col = new int[numValidMoves];
		for(int i=0; i<numValidMoves; i++) {

			float score = evaluateMove(actualField, validMoves_row[i], validMoves_col[i]);

			if(score > bestScore) {
				numBestMoves = 1;
				bestMovesIndex = 0;
				bestMoves_row[0] = validMoves_row[i];
				bestMoves_col[0] = validMoves_col[i];
				bestScore = score;
			}
			else if(FISEQUAL(score, bestScore)) { // score == bestScore?
				numBestMoves++;
				bestMovesIndex++;
				bestMoves_row[bestMovesIndex] = validMoves_row[i];
				bestMoves_col[bestMovesIndex] = validMoves_col[i];
			}

		}

		/* Choose one of best moves randomly */
		int moveIndex = rand() % numBestMoves;
		*moveToRow = bestMoves_row[moveIndex];
		*moveToCol = bestMoves_col[moveIndex];

		/* Free memory */
		delete[] bestMoves_row;
		delete[] bestMoves_col;
		delete[] validMoves_row;
		delete[] validMoves_col;

	}


	/*
	 * Implementation of class RandomPlayer
	 */

	RandomPlayer::RandomPlayer(enum color team) : AtomvsPlayer(team) {
		srand((unsigned) time(NULL)); // initialize random number generator
	}

	void RandomPlayer::makeMove(const AtomicField* actualField, int* moveToRow, int* moveToCol) {

		/* Get valid moves */
		int numValidMoves = getNumValidMoves(actualField);
		int* validMoves_row = new int[numValidMoves];
		int* validMoves_col = new int[numValidMoves];
		getValidMoves(actualField, validMoves_row, validMoves_col);

		/* Choose move randomly */
		int moveIndex = rand() % numValidMoves;
		*moveToRow = validMoves_row[moveIndex];
		*moveToCol = validMoves_col[moveIndex];

		/* Free memory */
		delete[] validMoves_row;
		delete[] validMoves_col;

	}


	/*
	 * Implementation of class MaxFieldsPlayer
	 */

	MaxFieldsPlayer::MaxFieldsPlayer(enum color team) : EvalPlayer(team) { }

	float MaxFieldsPlayer::evaluateMove(const AtomicField* actualField, int row, int col) {
		float score;
		AtomicField evalField = *actualField; // copy actualField
		evalField.makeMove(row, col, this->team);
		score = (float) evalField.getScore(this->team);
		return score;
	}


	/*
	 * Implementation of class MinFieldsPlayer
	 */

	MinFieldsPlayer::MinFieldsPlayer(enum color team) : EvalPlayer(team) { }

	float MinFieldsPlayer::evaluateMove(const AtomicField* actualField, int row, int col) {
		float score;
		AtomicField evalField = *actualField; // copy actualField
		evalField.makeMove(row, col, this->team);
		if(this->team == RED) {
			score = (float) evalField.getScore(BLUE);
		}
		else if(this->team == BLUE) {
			score = (float) evalField.getScore(RED);
		}
		float maxScore = (float) actualField->getSize() * actualField->getSize() + 1;
		return maxScore - score;
	}


	/*
	 * Implementation of class MaxParticlesPlayer
	 */

	MaxParticlesPlayer::MaxParticlesPlayer(enum color team) : EvalPlayer(team) { }

	float MaxParticlesPlayer::evaluateMove(const AtomicField* actualField, int row, int col) {
		float score;
		AtomicField evalField = *actualField; // copy actualField
		evalField.makeMove(row, col, this->team);
		score = (float) evalField.getNumParticles(this->team);
		return score;
	}

}

