/* 
 * File   : sdl_textmenu.cpp
 * Version: 2008-06-06
 *
 * Copyright (C) 2008 Jochen Peters
 *
 * dev@emirp.net
 * http://www.emirp.net/
 * 
 * This file is part of FissionField and FissionField2x.
 *
 * FissionField(2x) is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * FissionField(2x) is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with FissionField(2x). If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <iostream>
#include <SDL/SDL.h>
#include "SFont.h"
#include "sdl_textmenu.hpp"


sdlTextMenu::sdlTextMenu(SDL_Surface* _screen, int _numItems, int _activeIndex, const char* _fontFile, Uint8 _activeR, Uint8 _activeG, Uint8 _activeB, Uint8 _inactiveR, Uint8 _inactiveG, Uint8 _inactiveB) {

	pScreen = _screen;
	numItems = _numItems;
	activeIndex = _activeIndex;

	/* Set up menu items */
	items = new sdlTextMenuItem[numItems];
	for(int i=0; i<numItems; i++) {
		items[i].text[0] = '\0';
		items[i].x = 0;
		items[i].y = 0;
		items[i].w = 0;
		items[i].h = 0;
	}

	/* Set text colors */
	this->inactiveColor = SDL_MapRGB(_screen->format, _inactiveR, _inactiveG, _inactiveB);
	this->activeColor = SDL_MapRGB(_screen->format, _activeR, _activeG, _activeB);

	/* Load font */
	SDL_Surface* fontBitmap = SDL_LoadBMP(_fontFile);
	font = SFont_InitFont(fontBitmap);
	if(!font) {
		std::cerr << "An error occured while loading font. Abort." << std::endl;
		exit(1);
	}
	
}


sdlTextMenu::~sdlTextMenu() {
	delete[] items;
	SFont_FreeFont(font);
}


void sdlTextMenu::setItem(int _index, char* _text, int _posX, int _posY) {
	memcpy(items[_index].text, _text, SDL_TEXTMENU_MAXLENGTH);
	items[_index].x = _posX;
	items[_index].y = _posY;
	items[_index].w = SFont_TextWidth(font, _text);
	items[_index].h = SFont_TextHeight(font);
}


void sdlTextMenu::updateItem(int _index, char* _text, int _posX, int _posY, Uint32 _bgcolor) {
	if(_index >= 0 && _index < numItems) {
		if(items[_index].w > 0) {

			SDL_Rect area;
			area.x = items[_index].x;
			area.y = items[_index].y;
			area.w = items[_index].w;
			area.h = items[_index].h;

			clearItem(_index, _bgcolor);
			setItem(_index, _text, _posX, _posY);
			displayItem(_index);
		
			SDL_UpdateRect(pScreen, area.x, area.y, area.w, area.h);

		}
	}
}

void sdlTextMenu::updateItem(int _index, char* _text, Uint32 _bgcolor) {
	updateItem(_index, _text, items[_index].x, items[_index].y, _bgcolor);
}

void sdlTextMenu::clearItem(int _index, Uint32 _bgcolor) {
	if(_index >= 0 && _index < numItems) {
		if(items[_index].w > 0) {

			SDL_Rect area;
			area.x = items[_index].x;
			area.y = items[_index].y;
			area.w = items[_index].w;
			area.h = items[_index].h;

			SDL_FillRect(pScreen, &area, _bgcolor);

		}
	}
}

void sdlTextMenu::displayItem(int _index) {
	if(_index >= 0 && _index < numItems) {
		if(items[_index].w > 0) {

			SFont_Write(pScreen, font, items[_index].x, items[_index].y, items[_index].text);

			if(_index == activeIndex) {
				activate(activeIndex);
			}

		}
	}
}


void sdlTextMenu::displayMenu() {
	for(int i=0; i<numItems; i++) {
		displayItem(i);
	}
	//activate(activeIndex);
}


int sdlTextMenu::getActiveIndex() {
	return activeIndex;
}


void sdlTextMenu::activate(int _index) {
	if(_index >= 0 && _index < numItems) {
		if(items[_index].w > 0) {

			colorItem(activeIndex, false);
			activeIndex = _index;
			colorItem(activeIndex, true);

		}
	}
}


void sdlTextMenu::activatePrev() {
	if(activeIndex > 0) {
		activate(activeIndex-1);
	}
}


void sdlTextMenu::activateNext() {
	if(activeIndex < numItems-1) {
		activate(activeIndex+1);
	}
}


void sdlTextMenu::colorItem(int _index, bool _active) {

	if(_index >= 0 && _index < numItems) {
		if(items[_index].w > 0) {
	
			SDL_Rect area;
			area.x = items[_index].x;
			area.y = items[_index].y;
			area.w = items[_index].w;
			area.h = items[_index].h;
		
			if(_active) {
				replaceColor(pScreen, &area, inactiveColor, activeColor);
			}
			else {
				replaceColor(pScreen, &area, activeColor, inactiveColor);
			}
		
			SDL_UpdateRect(pScreen, area.x, area.y, area.w, area.h);

		}
	}
}


void sdlTextMenu::replaceColor(SDL_Surface* _surf, SDL_Rect* _area, Uint32 _fromColor, Uint32 _toColor) {

	if(SDL_MUSTLOCK(_surf)) {
		SDL_LockSurface(_surf);
	}

	unsigned int* ptr = (unsigned int*) _surf->pixels;
	for(int xpos=_area->x; xpos < _area->x + _area->w; xpos++) {
		for(int ypos=_area->y; ypos < _area->y + _area->h; ypos++) {
			
			switch(_surf->format->BytesPerPixel) {

				case 1: { // 8 bit per pixel
					Uint8* bufp = (Uint8*)_surf->pixels + ypos*_surf->pitch + xpos; // 8bit color!
					if(*bufp == _fromColor) {
						*bufp = _toColor;
					}
				}
				break;

				case 2: { // 15 or 16 bit per pixel
					Uint16* bufp = (Uint16*)_surf->pixels + ypos*_surf->pitch/2 + xpos; // 16bit color!
					if(*bufp == _fromColor) {
						*bufp = _toColor;
					}
				}
				break;

				case 3: { // 24 bit per pixel (slow)
					// NOT SUPPORTED!
					// do nothing
				}
				break;

				case 4: { // 32 bit per pixel
					Uint32* bufp = (Uint32*)_surf->pixels + ypos*_surf->pitch/4 + xpos; // 32bit color!
					if(*bufp == _fromColor) {
						*bufp = _toColor;
					}
				}
				break;

			}
		}
	}

	if(SDL_MUSTLOCK(_surf)) {
		SDL_UnlockSurface(_surf);
	}

}

